//
// JTypist v3.1 - Java typing tutor program for UNIX systems
// Copyright (C) 1999  Simon Baldwin (simonb@sco.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//
// KBLayout class - defines a keyboard layout for the given locale
//
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;

/**
 * Defines a keyboard layout for the given locale.
 */
public abstract class KBLayout {

	// description and keymap table; we use a vector since the ordering
	// of entries is used in ordering the statistics, and we want to
	// preserve it
	private String		kb_id			= null;
	private Vector		keymap_table		= new Vector();

	// lookup tables of glyphs by character they represent, and by keycode
	// they are assigned to, separated into shifted and unshifted
	private Hashtable	glyphs_by_char		= new Hashtable();
	private Hashtable	glyphs_by_keycode_unshifted
							= new Hashtable();
	private Hashtable	glyphs_by_keycode_shifted
							= new Hashtable();


	/*
	 * Accessor methods.
	 */
	public String	get_kb_id()	{ return ( kb_id ); }
	public void	set_kb_id( String kb_id ) {
		this.kb_id = kb_id;
	}
	public Enumeration
			get_elements()	{ return ( keymap_table.elements() ); }


	/**
	 * Method to add a keymap entry into the table.
	 */
	public void append( Glyph glyph ) {

		// add the entry to the main list
		keymap_table.addElement( glyph );

		// add the entry to the lookup tables
		glyphs_by_char.put( glyph.get_character(), glyph );
		if ( ! glyph.is_shifted() )
			glyphs_by_keycode_unshifted.put(
				new Integer( glyph.get_keycode()), glyph );
		else
			glyphs_by_keycode_shifted.put(
				new Integer( glyph.get_keycode()), glyph );
	}

	/**
	 * Method to replace an existing keymap entry in the table with
	 * another.   The existing entry is a glyph having the same
	 * character.  If none exists, the glyph is added to the table.
	 */
	public void replace( Glyph glyph ) {

		// note the character for this glyph
		Character glyph_character	= glyph.get_character();

		// search for any existing glyphs that implement the same
		// keyboard character
		Glyph old_glyph = (Glyph)(glyphs_by_char.get( glyph_character ));
		if ( old_glyph != null ) {

			// replace the old one with the new one, at the
			// same point in the original list
			int index = keymap_table.indexOf( old_glyph );
			keymap_table.setElementAt( glyph, index );

			// update the lookup tables
			glyphs_by_char.put( glyph.get_character(), glyph );
			if ( ! glyph.is_shifted() )
				glyphs_by_keycode_unshifted.put( new Integer(
						glyph.get_keycode()), glyph );
			else
				glyphs_by_keycode_shifted.put( new Integer(
						glyph.get_keycode()), glyph );

			// stop there
			return;
		}

		// nothing to replace, so just append instead
		append( glyph );
	}

	/**
	 * Method to delete a glyph from the keymap table.
	 */
	public void delete( char character ) {

		// search for any existing glyphs that implement the character
		Glyph old_glyph = (Glyph)(glyphs_by_char.get(
						new Character( character )));
		if ( old_glyph != null ) {

			// remove the glyph from the table
			keymap_table.removeElement( old_glyph );

			// remove the glyph from the lookup tables
			glyphs_by_char.remove( old_glyph.get_character() );
			if ( ! old_glyph.is_shifted() )
				glyphs_by_keycode_unshifted.remove( new Integer(
						old_glyph.get_keycode() ));
			else
				glyphs_by_keycode_shifted.remove( new Integer(
						old_glyph.get_keycode() ));
		}
	}


	/**
	 * Find the glyph that represents a given character.
	 */
	public Glyph find_glyph( char character ) {

		// return a simple lookup
		return ( (Glyph)(glyphs_by_char.get(
						new Character( character ))) );
	}

	/**
	 * Find the glyph that represents a given keycode, with shift state.
	 */
	public Glyph find_glyph( int keycode, boolean shifted ) {

		// return a simple lookup
		if ( ! shifted )
			return ( (Glyph)(glyphs_by_keycode_unshifted.get(
						new Integer( keycode ))) );
		else
			return ( (Glyph)(glyphs_by_keycode_shifted.get(
						new Integer( keycode ))) );
	}


	/**
	 * Printout method.
	 */
	public String toString() {
		String ret_string = getClass().getName() + "{"
				+ "kb_id=\"" + kb_id
				+ "\",keymap_table@" + keymap_table;
		return ( ret_string + "}" );
	}
}
