//
// JTypist v3.1 - Java typing tutor program for UNIX systems
// Copyright (C) 1999  Simon Baldwin (simonb@sco.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//
// KeyboardGUI class - creates a frame to display an animated keyboard
//
import java.awt.*;
import java.awt.event.*;
import java.applet.*;
import java.util.Hashtable;
import java.util.Enumeration;

/**
 * Displays an animated keyboard graphic.
 */
public class KeyboardGUI extends Frame
			implements Runnable,
			ItemListener, ActionListener, WindowListener {
	private boolean inAnApplet = true;

	// menubar GUI components
	private MenuBar			menu_bar;
	private Menu			file_menu;
	private MenuItem		file_close;
	private Menu			options_menu;
	private CheckboxMenuItem	options_expected, options_colourize,
					options_error, options_success;

	// key keycodes and ranges; note - when this module and others talk
	// about keycodes, these are really only a convenient way to assign
	// unique numbers to keyboard positions; although the values used
	// are real keycodes, they could have been any unique scheme
	private final int	ROW_1_1			=  41;
	private final int	ROW_1_2_START		=   2;
	private final int	ROW_1_2_END		=  13;
	private final int	ROW_1_BACKSPACE		=  14;
	private final int	ROW_2_TAB		=  15;
	private final int	ROW_2_1_START		=  16;
	private final int	ROW_2_1_END		=  27;
	private final int	ROW_2_2			=  43;
	private final int	ROW_3_CAPSLOCK		=  58;
	private final int	ROW_3_1_START		=  30;
	private final int	ROW_3_1_END		=  40;
	private final int	ROW_3_ENTER		=  28;
	private final int	ROW_4_LEFT_SHIFT	=  42;
	private final int	ROW_4_1_START		=  44;
	private final int	ROW_4_1_END		=  53;
	private final int	ROW_4_RIGHT_SHIFT	=  54;
	private final int	ROW_5_LEFT_CTRL		=  29;
	private final int	ROW_5_LEFT_ALT		=  56;
	private final int	ROW_5_SPACE		=  57;
	private final int	ROW_5_RIGHT_ALT		= 100;
	private final int	ROW_5_RIGHT_CTRL	=  97;

	// keycodes for numeric keypad keys
	private final int	NUM_NUMLOCK		=  69;
	private final int	NUM_SLASH		=  98;
	private final int	NUM_ASTERISK		=  55;
	private final int	NUM_MINUS		=  74;
	private final int	NUM_7			=  71;
	private final int	NUM_8			=  72;
	private final int	NUM_9			=  73;
	private final int	NUM_4			=  75;
	private final int	NUM_5			=  76;
	private final int	NUM_6			=  77;
	private final int	NUM_PLUS		=  78;
	private final int	NUM_1			=  79;
	private final int	NUM_2			=  80;
	private final int	NUM_3			=  81;
	private final int	NUM_0			=  82;
	private final int	NUM_PERIOD		=  83;
	private final int	NUM_ENTER		=  96;

	// standard colours for various conditions
	private final Color	COLOR_NEUTRAL		= Color.lightGray;
	private final Color	COLOR_EXPECTED		= Color.black;
	private final Color	COLOR_SUCCESS		= Color.green;
	private final Color	COLOR_ERROR		= Color.red;

	// associate some colours with fingers
	private final Color	COLOUR_LH_PINKY		= Color.pink;

	// a note of the keyboard layout we are monitoring
	private KBLayout	kblayout		= null;

	// tables of keyicons by keycode, and fingericons by finger
	private Hashtable	keyicons_by_keycode	= new Hashtable();
	private Hashtable	fingericons_by_finger	= new Hashtable();

	// table of colours by finger, for associations when learning
	private Hashtable	colours_by_finger	= new Hashtable();

	// additional table of keyicons for the numeric keypad
	private Hashtable	numkeycodes_by_char	= new Hashtable();

	// keyicons that are currently coloured up in some way
	private KeyIcon		last_expected		= null;
	private KeyIcon		num_last_expected	= null;
	private KeyIcon		last_error		= null;
	private KeyIcon		num_last_error		= null;
	private KeyIcon		last_success		= null;
	private KeyIcon		num_last_success	= null;

	// timestamps on the last error and success highlights
	private long		timestamp_error		= 0;
	private long		timestamp_success	= 0;

	// finger for a fingericon that is currently coloured up
	private Finger		last_expected_finger	= null;

	// semi-constant keyicons, for shift keys
	private KeyIcon		left_shift		= null;
	private KeyIcon		right_shift		= null;


	/**
	 * Initialize the whole GUI set.
	 */
	public KeyboardGUI( boolean inAnApplet, KBLayout kblayout ) {
		this.inAnApplet = inAnApplet;

		// note our kblayout
		this.kblayout = kblayout;

		// create GUI components
		setTitle( "Typist v3.1 Keyboard Window" );
		setLayout( new BorderLayout() );

		// create the menu bar stuff
		menu_bar = new MenuBar();
		setMenuBar( menu_bar );
		// File menu
		file_menu = new Menu( "File" );
		menu_bar.add( file_menu );
		file_close = new MenuItem( "Close window",
				new MenuShortcut( KeyEvent.VK_C ));
		file_close.addActionListener( this );
		file_menu.add( file_close );
		// Option menu
		options_menu = new Menu( "Options" );
		menu_bar.add( options_menu );
		options_expected = new CheckboxMenuItem(
		                "Show exercise key fingering", true );
		options_expected.addItemListener( this );
		options_menu.add( options_expected );
		options_colourize = new CheckboxMenuItem(
		                "Show fingering with colours", true );
		options_colourize.addItemListener( this );
		options_menu.add( options_colourize );
		options_error = new CheckboxMenuItem(
		                "Show keying errors", true );
		options_error.addItemListener( this );
		options_menu.add( options_error );
		options_success = new CheckboxMenuItem(
		                "Show correct keystrokes", false );
		options_success.addItemListener( this );
		options_menu.add( options_success );

		// set up the associations between fingers and colours;
		// unspecified ones are left neutral
		colours_by_finger.put( Finger.LH_PINKY,  Color.pink );
		colours_by_finger.put( Finger.LH_RING,   Color.orange );
		colours_by_finger.put( Finger.LH_MIDDLE, Color.cyan );
		colours_by_finger.put( Finger.LH_INDEX,  Color.yellow );
		colours_by_finger.put( Finger.RH_INDEX,  Color.pink );
		colours_by_finger.put( Finger.RH_MIDDLE, Color.orange );
		colours_by_finger.put( Finger.RH_RING,   Color.cyan );
		colours_by_finger.put( Finger.RH_PINKY,  Color.yellow );

		// create the graphics display with initially null legends
		Panel panel = make_kb_panel();
		panel.setBackground( Color.white );

		// add keyicon legends to the display
		add_keyicon_legends();

		// add the graphics display to the main panel
		add( "Center", panel );

		// create the keyboard type indicator label
		Label kbtype_label = new Label( "Keyboard type: "
						+ kblayout.get_kb_id(),
				Label.RIGHT );
		add( "South", kbtype_label );

		// set up left and right shift keyicons to save many
		// lookups later
		left_shift	= (KeyIcon)(keyicons_by_keycode.get(
					new Integer( ROW_4_LEFT_SHIFT )));
		right_shift	= (KeyIcon)(keyicons_by_keycode.get(
					new Integer( ROW_4_RIGHT_SHIFT )));

		// set up the numeric keypad table, so that certain characters
		// get shown on that keypad as well; note that this is pretty
		// static - we just assume the numeric keypad doesn't vary
		// for different keyboard types
		numkeycodes_by_char.put( new Character( '/' ),
						 new Integer( NUM_SLASH ) );
		numkeycodes_by_char.put( new Character( '*' ),
						 new Integer( NUM_ASTERISK ) );
		numkeycodes_by_char.put( new Character( '-' ),
						 new Integer( NUM_MINUS ) );
		numkeycodes_by_char.put( new Character( '7' ),
						 new Integer( NUM_7 ) );
		numkeycodes_by_char.put( new Character( '8' ),
						 new Integer( NUM_8 ) );
		numkeycodes_by_char.put( new Character( '9' ),
						 new Integer( NUM_9 ) );
		numkeycodes_by_char.put( new Character( '4' ),
						 new Integer( NUM_4 ) );
		numkeycodes_by_char.put( new Character( '5' ),
						 new Integer( NUM_5 ) );
		numkeycodes_by_char.put( new Character( '6' ),
						 new Integer( NUM_6 ) );
		numkeycodes_by_char.put( new Character( '+' ),
						 new Integer( NUM_PLUS ) );
		numkeycodes_by_char.put( new Character( '1' ),
						 new Integer( NUM_1 ) );
		numkeycodes_by_char.put( new Character( '2' ),
						 new Integer( NUM_2 ) );
		numkeycodes_by_char.put( new Character( '3' ),
						 new Integer( NUM_3 ) );
		numkeycodes_by_char.put( new Character( '0' ),
						 new Integer( NUM_0 ) );
		numkeycodes_by_char.put( new Character( '.' ),
						 new Integer( NUM_PERIOD ) );
		numkeycodes_by_char.put( new Character( '\n' ),
						 new Integer( NUM_ENTER ) );

		// colourize the keyboard display
		colourize_keyicons();
		colourize_fingericons();

		// start the timeout thread
		new Thread( this, "KeyboardGUI_timeout" ).start();
	}


	/**
	 * Thread to automatically call refresh, if the option to do it
	 * is set.
	 */
	public void run() {

		// thread 'tick' period, in mS, and error/success timeouts
		final int	THREAD_TICK		= 250;
		final int	HIGHLIGHT_TIMEOUT	= 1000;

		// loop forever, at the lowest priority
		Thread.currentThread().setPriority( Thread.MIN_PRIORITY );
		while ( true ) {

			// wait for the next tick, and ignore it if our
			// window isn't visible
			do {
				try { Thread.sleep( THREAD_TICK ); }
				catch ( InterruptedException e ) {}
			} while ( ! isVisible() );

			// get the current system time
			long now = System.currentTimeMillis();

			// if the error highlight has been up long enough then
			// drop it
			if ( last_error != null
					&& now - timestamp_error >=
							HIGHLIGHT_TIMEOUT )
				clear_error();

			// do the same for then success highlight
			if ( last_success != null
					&& now - timestamp_success >=
							HIGHLIGHT_TIMEOUT )
				clear_success();
		}
	}


	/*
	 * Create the complete keyboard display panel.
	 */
	private Panel make_kb_panel() {

		// create the overall panel, and set six rows
		Panel panel = new Panel();
		panel.setLayout( new GridLayout( 6, 0, 0, 0 ));

		// create the top row of keys, all blank for now
		Panel row_1 = new Panel();
		row_1.setLayout( new FlowLayout( FlowLayout.CENTER, 0, 0 ) );
		row_1.add( new_keyicon( "", 100, ROW_1_1 ));
		for ( int keycode = ROW_1_2_START; keycode <= ROW_1_2_END;
				keycode++ )
			row_1.add( new_keyicon( "", 100, keycode ));
		row_1.add( new_keyicon( "Bkspc", 200, ROW_1_BACKSPACE ));

		// create the second row of keys, again all blank
		Panel row_2 = new Panel();
		row_2.setLayout( new FlowLayout( FlowLayout.CENTER, 0, 0 ) );
		row_2.add( new_keyicon( "Tab", 150, ROW_2_TAB ));
		for ( int keycode = ROW_2_1_START; keycode <= ROW_2_1_END;
				keycode++ )
			row_2.add( new_keyicon( "", 100, keycode ));
		row_2.add( new_keyicon( "", 150, ROW_2_2 ));

		// the third row
		Panel row_3 = new Panel();
		row_3.setLayout( new FlowLayout( FlowLayout.CENTER, 0, 0 ) );
		row_3.add( new_keyicon( "Caps", 175, ROW_3_CAPSLOCK ));
		for ( int keycode = ROW_3_1_START; keycode <= ROW_3_1_END;
				keycode++ )
			row_3.add( new_keyicon( "", 100, keycode ));
		row_3.add( new_keyicon( "Enter", 225, ROW_3_ENTER ));

		// the fourth row
		Panel row_4 = new Panel();
		row_4.setLayout( new FlowLayout( FlowLayout.CENTER, 0, 0 ) );
		row_4.add( new_keyicon( "Shift", 225, ROW_4_LEFT_SHIFT ));
		for ( int keycode = ROW_4_1_START; keycode <= ROW_4_1_END;
				keycode++ )
			row_4.add( new_keyicon( "", 100, keycode ));
		row_4.add( new_keyicon( "Shift", 275, ROW_4_RIGHT_SHIFT ));

		// the bottom keyboard row
		Panel row_5 = new Panel();
		row_5.setLayout( new FlowLayout( FlowLayout.CENTER, 0, 0 ) );
		row_5.add( new_keyicon( "Ctrl", 150, ROW_5_LEFT_CTRL ));
		row_5.add( new_keyicon( null,   100, -1 ));	// spacer
		row_5.add( new_keyicon( "Alt",  150, ROW_5_LEFT_ALT ));
		row_5.add( new_keyicon( "",     700, ROW_5_SPACE ));
		row_5.add( new_keyicon( "Alt",  150, ROW_5_RIGHT_ALT ));
		row_5.add( new_keyicon( null,   100, -1 ));	// spacer
		row_5.add( new_keyicon( "Ctrl", 150, ROW_5_RIGHT_CTRL ));

		// keypad additions to the above rows
		row_1.add( new_keyicon( null,  200, -1 ));	// spacer
		row_1.add( new_keyicon( "Nm",  100, NUM_NUMLOCK ));
		row_1.add( new_keyicon( "/",   100, NUM_SLASH ));
		row_1.add( new_keyicon( "*",   100, NUM_ASTERISK ));
		row_1.add( new_keyicon( "-",   100, NUM_MINUS ));
		row_2.add( new_keyicon( null,  200, -1 ));	// spacer
		row_2.add( new_keyicon( "7",   100, NUM_7 ));
		row_2.add( new_keyicon( "8",   100, NUM_8 ));
		row_2.add( new_keyicon( "9",   100, NUM_9 ));
		row_2.add( new_keyicon( "",    100, -1 ));	// spacer
		row_3.add( new_keyicon( null,  200, -1 ));	// spacer
		row_3.add( new_keyicon( "4",   100, NUM_4 ));
		row_3.add( new_keyicon( "5",   100, NUM_5 ));
		row_3.add( new_keyicon( "6",   100, NUM_6 ));
		row_3.add( new_keyicon( "+",   100, NUM_PLUS ));
		row_4.add( new_keyicon( null,  200, -1 ));	// spacer
		row_4.add( new_keyicon( "1",   100, NUM_1 ));
		row_4.add( new_keyicon( "2",   100, NUM_2 ));
		row_4.add( new_keyicon( "3",   100, NUM_3 ));
		row_4.add( new_keyicon( "",    100, -1 ));	// spacer
		row_5.add( new_keyicon( null,  200, -1 ));	// spacer
		row_5.add( new_keyicon( "0",   200, NUM_4 ));
		row_5.add( new_keyicon( ".",   100, NUM_PERIOD ));
		row_5.add( new_keyicon( "En",  100, NUM_ENTER ));

		// the finger icons row
		Panel left_finger_row	= new Panel();
		Panel right_finger_row	= new Panel();
		left_finger_row.setLayout( new FlowLayout() );
		right_finger_row.setLayout( new FlowLayout() );
		left_finger_row.add(  new_fingericon( "L", Finger.LH_PINKY ));
		left_finger_row.add(  new_fingericon( "R", Finger.LH_RING ));
		left_finger_row.add(  new_fingericon( "M", Finger.LH_MIDDLE ));
		left_finger_row.add(  new_fingericon( "I", Finger.LH_INDEX ));
		left_finger_row.add(  new_fingericon( "T", Finger.LH_THUMB ));
		right_finger_row.add( new_fingericon( "T", Finger.RH_THUMB ));
		right_finger_row.add( new_fingericon( "I", Finger.RH_INDEX ));
		right_finger_row.add( new_fingericon( "M", Finger.RH_MIDDLE ));
		right_finger_row.add( new_fingericon( "R", Finger.RH_RING ));
		right_finger_row.add( new_fingericon( "L", Finger.RH_PINKY ));
		Panel finger_row = new Panel();
		finger_row.setLayout( new GridLayout( 0, 3 ));
		finger_row.add( left_finger_row );
		finger_row.add( right_finger_row );
		finger_row.add( new Panel() );			// spacer

		// add the rows to the main panel
		panel.add( row_1 ); panel.add( row_2 ); panel.add( row_3 );
		panel.add( row_4 ); panel.add( row_5 ); panel.add( finger_row );

		// return the main panel to the caller
		return ( panel );
	}

	/*
	 * Add a new keyicon with the given legend and width factor,
	 * for the given keycode.
	 */
	private KeyIcon new_keyicon( String legend, int width_factor,
			int keycode ) {

		// create the new key icon
		KeyIcon keyicon = new KeyIcon( legend, width_factor );
		keyicon.set_colour( COLOR_NEUTRAL );

		// add it to the table for this keycode
		keyicons_by_keycode.put( new Integer( keycode ), keyicon );

		// return it so it can be added to the graphics
		return( keyicon );
	}

	/*
	 * Add a new finger with the given legend and finger.
	 */
	private FingerIcon new_fingericon( String legend, Finger finger ) {

		// create the new finger icon
		FingerIcon fingericon = new FingerIcon( legend );
		fingericon.set_colour( COLOR_NEUTRAL );

		// add it to the table for this keycode
		fingericons_by_finger.put( finger, fingericon );

		// return it so it can be added to the graphics
		return( fingericon );
	}


	/**
	 * Add legends to keyicons represented in the keyboard layout, and
	 * set background colours to indicate assigned fingers.
	 */
	private void add_keyicon_legends() {

		// scan the keymap, setting legends for everything represented
		for ( Enumeration enum = kblayout.get_elements();
				enum.hasMoreElements(); ) {
			Glyph glyph = (Glyph)(enum.nextElement());

			// note the keycode, glyph character, and shift state
			int keycode		= glyph.get_keycode();
			Character character	= glyph.get_character();
			boolean shifted		= glyph.is_shifted();

			// find the keyicon that does this keycode
			KeyIcon keyicon = (KeyIcon)(keyicons_by_keycode.get(
						new Integer( keycode )));

			if ( keyicon != null ) {

				// get the current legend
				String current_legend = keyicon.get_legend();

				// if it is null, blank, or is not but it is an
				// unshifted glyph, then retrofit the keyicon
				// with a legend - this is somewhat heuristic,
				// since the kblayout doesn't actually tell us
				// precisely what to use as a legend
				if ( current_legend == null
						|| current_legend.equals( "" )
						|| ! shifted ) {

					// create a legend for the glyph
					if ( character.charValue() == '\n' ||
						character.charValue() == ' ' )
						keyicon.set_legend(
							glyph.get_text() );
					else
						keyicon.set_legend( ( "" +
							character.toString() ));
				}
			}
		}
	}


	/**
	 * Colourize the background of keyicons to show the associations
	 * between fingers and keys, depending on the option setting.
	 */
	private void colourize_keyicons() {

		// note the state of the colourizing option
		boolean do_colours = options_colourize.getState();

		// scan the keymap, setting colours for everything represented
		for ( Enumeration enum = kblayout.get_elements();
				enum.hasMoreElements(); ) {
			Glyph glyph = (Glyph)(enum.nextElement());

			// note the keycode and finger
			int keycode		= glyph.get_keycode();
			Finger finger		= glyph.get_finger();

			// find the keyicon that does this keycode
			KeyIcon keyicon = (KeyIcon)(keyicons_by_keycode.get(
						new Integer( keycode )));
			if ( keyicon != null ) {

				// set initially to match the main panel
				keyicon.set_bgcolour( Color.white );

				// should we do colours?
				if ( do_colours ) {
					// colour-code the keyicon for the
					// finger, if it has an association
					Color colour = (Color)
						(colours_by_finger.get( finger ));
					if ( colour != null )
						keyicon.set_bgcolour( colour );
				}
			}
		}
	}

	/**
	 * Colourize the background of fingericons to show the associations
	 * between fingers and keys, depending on the option setting.
	 */
	private void colourize_fingericons() {

		// note the state of the colourizing option
		boolean do_colours = options_colourize.getState();

		// loop through every finger known
		for ( Enumeration enum = Finger.get_elements();
				enum.hasMoreElements(); ) {
			Finger finger = (Finger)(enum.nextElement());

			// find the fingericon that does this finger
			FingerIcon fingericon = (FingerIcon)
					(fingericons_by_finger.get( finger ));
			if ( fingericon != null ) {

				// set a neutral colour initially
				fingericon.set_colour( COLOR_NEUTRAL );

				// should we do colours?
				if ( do_colours ) {
					// colour-code the keyicon for the
					// finger, if it has an association
					Color colour = (Color)
						(colours_by_finger.get( finger ));
					if ( colour != null )
						fingericon.set_colour( colour );
				}
			}
		}
	}


	/*
	 * Set the keyicon for a given keycode, glyph, or character
	 * to a given colour.
	 */
	private KeyIcon set_keyicon_colour( Integer keycode, Color colour ) {

		// look up the keyicon for this keycode
		KeyIcon keyicon = (KeyIcon)(keyicons_by_keycode.get( keycode ));
		if ( keyicon != null ) {

			// set the keyicon colour
			keyicon.set_colour( colour );
			return ( keyicon );
		}
		return ( null );
	}
	private KeyIcon set_keyicon_colour( int keycode, Color colour ) {

		// call the Integer version
		return ( set_keyicon_colour( new Integer( keycode ), colour ));
	}
	private KeyIcon set_keyicon_colour( Glyph glyph, Color colour ) {

		// call the int version
		return ( set_keyicon_colour( glyph.get_keycode(), colour ));
	}
	private KeyIcon set_keyicon_colour( char character, Color colour ) {

		// look up the glyph for this character
		Glyph glyph = kblayout.find_glyph( character );
		if ( glyph != null )
			return ( set_keyicon_colour( glyph, colour ));
		return ( null );
	}
	private KeyIcon num_set_keyicon_colour( char character, Color colour ) {

		// see if this key could also be on the numeric keypad
		Integer num_keycode = (Integer)(numkeycodes_by_char.get(
						new Character( character )));
		if ( num_keycode != null )
			return ( set_keyicon_colour( num_keycode, colour ));
		return ( null );
	}


	/**
	 * Indicate a keyicon and fingericon as the next key expected.
	 */
	public void char_expected( char character ) {

		// ignore the call if the option is not set
		if ( options_expected.getState() ) {

			// clear any previous expectations
			clear_expected();

			// get the glyph for this character
			Glyph glyph = kblayout.find_glyph( character );
			if ( glyph != null ) {

				// identify the fingericon for this glyph
				Finger finger		= glyph.get_finger();
				FingerIcon fingericon	= (FingerIcon)
					(fingericons_by_finger.get( finger ));
				if ( fingericon != null )
					fingericon.set_colour( COLOR_EXPECTED );
				last_expected_finger	= finger;

				// if a shift is required with this char then
				// highlight that key too
				if ( glyph.is_shifted()
						&& finger.is_left_handed() )
					right_shift.set_colour( COLOR_EXPECTED );
				if ( glyph.is_shifted()
						&& finger.is_right_handed() )
					left_shift.set_colour( COLOR_EXPECTED );

				// set a main keyboard keyicon to a new colour
				last_expected = set_keyicon_colour( glyph,
							COLOR_EXPECTED );
			}

			// independent of all of that, try setting a keyicon one
			// the numeric keypad
			num_last_expected = num_set_keyicon_colour( character,
							COLOR_EXPECTED );
		}
	}

	/**
	 * Clear keyicon and fingericon expectations.
	 */
	public void clear_expected() {

		// reset the last keyicons to background
		if ( last_expected != null )
			last_expected.set_colour( COLOR_NEUTRAL );
		last_expected = null;
		if ( num_last_expected != null )
			num_last_expected.set_colour( COLOR_NEUTRAL );
		num_last_expected = null;

		// clear both shift keys
		left_shift.set_colour( COLOR_NEUTRAL );
		right_shift.set_colour( COLOR_NEUTRAL );

		// clear the last finger shown up
		if ( last_expected_finger != null ) {

			// get the fingericon for this finger
			FingerIcon fingericon = (FingerIcon)
					(fingericons_by_finger.get(
							last_expected_finger ));
			if ( fingericon != null ) {

				// set to neutral, or colourize as required
				fingericon.set_colour( COLOR_NEUTRAL );
				if ( options_colourize.getState() ) {

					Color colour = (Color)
						(colours_by_finger.get(
							last_expected_finger ));
					if ( colour != null )
						fingericon.set_colour( colour );
				}
			}
		}
		last_expected_finger = null;
	}

	/**
	 * Indicate a keyicon as pressed in error.  This highlighting will
	 * automatically timeout.
	 */
	public void char_error( char character ) {

		// ignore the call if the option is not set
		if ( options_error.getState() ) {

			// clear any previous errors
			clear_error();

			// set the right keyicon(s) to a new colour
			last_error	= set_keyicon_colour( character,
								COLOR_ERROR );
			num_last_error	= num_set_keyicon_colour( character,
								COLOR_ERROR );

			// note the time, for timeouts
			timestamp_error = System.currentTimeMillis();
		}
	}

	/**
	 * Clear keyicon error.  If the error keyicon is also the current
	 * expected keyicon, this will set it to indicate the expectation,
	 * otherwise it is returned to a neutral colour.
	 */
	public void clear_error() {

		// reset the last keyicons to background
		if ( last_error != null )
			if ( last_error != last_expected )
				last_error.set_colour( COLOR_NEUTRAL );
			else
				last_error.set_colour( COLOR_EXPECTED );
		last_error = null;
		if ( num_last_error != null )
			if ( num_last_error != num_last_expected )
				num_last_error.set_colour( COLOR_NEUTRAL );
			else
				num_last_error.set_colour( COLOR_EXPECTED );
		num_last_error = null;
	}

	/**
	 * Indicate a keyicon as pressed correctly.  This highlighting will
	 * automatically timeout.
	 */
	public void char_success( char character ) {

		// ignore the call if the option is not set
		if ( options_success.getState() ) {

			// clear any previous success
			clear_success();

			// set the right keyicon(s) to a new colour
			last_success	= set_keyicon_colour( character,
								COLOR_SUCCESS );
			num_last_success= num_set_keyicon_colour( character,
								COLOR_SUCCESS );

			// if this is the current expected keyicon, forget
			// the expectation so the success highlight isn't lost;
			// this is a normal case - the expected key was pressed,
			// and the expectation moving on will otherwise erase
			// our highlighting
			if ( last_success == last_expected )
				last_expected = null;
			if ( num_last_success == num_last_expected )
				num_last_expected = null;

			// note the time, for timeouts
			timestamp_success = System.currentTimeMillis();
		}
	}

	/**
	 * Clear keyicon success.  If the success keyicon is also the current
	 * expected keyicon, this will set it to indicate the expectation,
	 * otherwise it is returned to a neutral colour.
	 */
	public void clear_success() {

		// reset the last keyicons to background
		if ( last_success != null )
			if ( last_success != last_expected )
				last_success.set_colour( COLOR_NEUTRAL );
			else
				last_success.set_colour( COLOR_EXPECTED );
		last_success = null;
		if ( num_last_success != null )
			if ( num_last_success != num_last_expected )
				num_last_success.set_colour( COLOR_NEUTRAL );
			else
				num_last_success.set_colour( COLOR_EXPECTED );
		num_last_success = null;
	}

	/*
	 * Handle dismiss, close, etc by just hiding ourselves.
	 */
	private void handle_close_button() {

		// simply just hide this frame from view
		setVisible( false );
	}

	/*
	 * Handle GUI component actions and events.
	 */
	public void actionPerformed( ActionEvent evt ) {
		if ( evt.getSource() instanceof MenuItem ) {
			if ( evt.getSource() == file_close )
				handle_close_button();
		}
	}
	public void windowClosing( WindowEvent evt ) {
		handle_close_button();
	}
	public void windowClosed( WindowEvent evt ) {}
	public void windowOpened( WindowEvent evt ) {}
	public void windowIconified( WindowEvent evt ) {}
	public void windowDeiconified( WindowEvent evt ) {}
	public void windowActivated( WindowEvent evt ) {}
	public void windowDeactivated( WindowEvent evt ) {}
	public void itemStateChanged( ItemEvent evt ) {
		if ( evt.getSource() instanceof CheckboxMenuItem ) {
			if ( evt.getSource() == options_expected )
				clear_expected();
			else if ( evt.getSource() == options_colourize ) {
				colourize_keyicons();
				colourize_fingericons();
			}
		}
	}

	/**
	 * Class test code.
	 */
	public static void main( String args[] ) {
		System.out.println( "*** CLASS TEST CODE ***" );
		KeyboardGUI kb = new KeyboardGUI( false, new KBLayoutUS() );
		kb.pack(); kb.setVisible( true );
		while ( true ) {
			kb.char_expected( (char)( Math.random() * 96 + 32 ) );
			try { Thread.sleep( 2000 ); }
			catch( InterruptedException e ) {}
			kb.char_error( (char)( Math.random() * 96 + 32 ) );
			try { Thread.sleep( 300 ); }
			catch( InterruptedException e ) {}
			kb.char_success( (char)( Math.random() * 96 + 32 ) );
			try { Thread.sleep( 300 ); }
			catch( InterruptedException e ) {}
		}
	}
}
