//
// JTypist v3.1 - Java typing tutor program for UNIX systems
// Copyright (C) 1999  Simon Baldwin (simonb@sco.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//
// StatisticsGUI class - creates a frame to display typing statistics
//
import java.awt.*;
import java.awt.event.*;
import java.applet.*;
import java.util.Vector;
import java.util.Enumeration;

/**
 * Creates a frame to display typing statistics.
 */
public class StatisticsGUI extends Frame
			implements Runnable,
			ItemListener, ActionListener, WindowListener {
	private boolean inAnApplet = true;

	// menubar GUI components
	private MenuBar			menu_bar;
	private Menu			file_menu;
	private MenuItem		file_refresh, file_reset, file_close;
	private Menu			view_menu;
	private CheckboxMenuItem	view_finger		= null;
	private Vector			view_glyphgroups	= new Vector();
	private Menu			options_menu;
	private CheckboxMenuItem	options_adjWPM;
	private CheckboxMenuItem	options_autorefresh;

	// GUI components
	private Panel		card_panel;
	private CardLayout	card_layout;
	private Button		refresh_button;
	private Vector		stats_descriptors	= new Vector();

	// number of gauges and labels on the panels we create
	private final int	COLUMNS_PER_STATISTIC	= 3;
	private final int	STATISTICS_PER_ROW	= 3;

	// somewhat arbitrary limits for warnings/errors on gauges; this
	// does the coloured gauges to try to help pick out the keys that
	// need more practice
	private final int	ERROR_ACCURACY	= 70;	// <70% is bad
	private final int	WARN_ACCURACY	= 90;	// <90% is a concern
	private final int	ERROR_SPEED	= 10;	// <10WPM is bad
	private final int	WARN_SPEED	= 30;	// <30WPM is a concern

	// statistics set we are working with, for reference
	private KBStatistics	stats_set	= null;

	// inner class tying together a set of statistics and the gauges
	// that represent it
	class StatsDescriptor {
		protected KeyingStatistics	key_stats	= null;
		protected Gauge			accuracy_gauge	= null;
		protected Gauge			speed_gauge	= null;
		protected StatsDescriptor( KeyingStatistics key_stats,
				Gauge accuracy_gauge, Gauge speed_gauge ) {
			this.key_stats		= key_stats;
			this.accuracy_gauge	= accuracy_gauge;
			this.speed_gauge	= speed_gauge;
		}
	}

	/**
	 * Initialize the whole GUI set.
	 */
	public StatisticsGUI( boolean inAnApplet, KBStatistics stats_set ) {
		this.addWindowListener( this );
		this.inAnApplet = inAnApplet;

		// create GUI components
		setTitle( "Typist v3.1 Typing Statistics Window" );
		setLayout( new BorderLayout() );

		// note the stats set for reference
		this.stats_set = stats_set;

		// create the menu bar stuff
		menu_bar = new MenuBar();
		setMenuBar( menu_bar );
		// File menu
		file_menu = new Menu( "File" );
		menu_bar.add( file_menu );
		file_refresh = new MenuItem( "Refresh statistics",
				new MenuShortcut( KeyEvent.VK_R ));
		file_menu.add( file_refresh );
		file_reset = new MenuItem( "Reset all",
				new MenuShortcut( KeyEvent.VK_A ));
		file_reset.addActionListener( this );
		file_menu.add( file_reset );
		file_close = new MenuItem( "Close window",
				new MenuShortcut( KeyEvent.VK_C ));
		file_close.addActionListener( this );
		file_menu.add( file_close );

		// first part of the view menu
		view_menu = new Menu( "View" );
		menu_bar.add( view_menu );

		// create the refresh button
		Panel toolbar = new Panel();
		toolbar.setLayout( new FlowLayout( FlowLayout.LEFT ));
		add( "North", toolbar );

		refresh_button = new Button( "Refresh" );
		refresh_button.addActionListener( this );
		toolbar.add( refresh_button );
		Label kbtype_label = new Label( "Keyboard type: "
				+ stats_set.get_kblayout().get_kb_id(),
				Label.RIGHT );
		add( "South", kbtype_label );

		// create a card panel for the stats to live on
		card_panel	= new Panel();
		card_layout	= new CardLayout();
		card_panel.setLayout( card_layout );

		// find the maximum number of statistics any panel
		// in the set has; this is used for padding the panels
		// with null labels so the sizes of all the gauges and
		// labels are equal
		int max_num_stats = -1;
		for ( Enumeration enum = GlyphGroup.get_elements();
				enum.hasMoreElements(); ) {
			GlyphGroup glyphgroup = (GlyphGroup)enum.nextElement();

			// see if this group has more statistics
			max_num_stats = Math.max(
					size_group_panel( stats_set,
							glyphgroup ),
					max_num_stats );
		}
		max_num_stats = Math.max(
				size_finger_panel( stats_set ),
				max_num_stats );

		// from this number, calculate the number of statistics,
		// real or padded, we want to put on each and every panel
		int stats_per_panel = ( (max_num_stats - 1) /
						STATISTICS_PER_ROW )
					* STATISTICS_PER_ROW
					+ STATISTICS_PER_ROW;

		// create the panel of finger statistics, and a menu item
		Panel fstats_panel = make_finger_stats_panel( stats_set,
				stats_per_panel );
		fstats_panel.setBackground( Color.white );
		view_finger = new CheckboxMenuItem( "Finger statistics" );
		view_finger.addItemListener( this );
		view_menu.add( view_finger );
		card_panel.add( view_finger.getLabel(), fstats_panel );

		// create a panel for each of the glyph groups defined,
		// and add a view menuitem so that each can be accessed
		for ( Enumeration enum = GlyphGroup.get_elements();
				enum.hasMoreElements(); ) {
			GlyphGroup glyphgroup = (GlyphGroup)enum.nextElement();

			// create a panel for this group
			Panel stats_panel = make_stats_panel( stats_set,
					glyphgroup, stats_per_panel );

			// see if there is anything on the panel
			if ( stats_panel != null ) {

				// get a string to identify this glyph group
				String glyphgroup_text = glyphgroup.get_text();

				// create a view menu item for this group, and
				// add it to the list we use when panels change
				CheckboxMenuItem cbmenuitem
						= new CheckboxMenuItem(
							glyphgroup_text );
				cbmenuitem.addItemListener( this );
				view_menu.add( cbmenuitem );
				view_glyphgroups.addElement( cbmenuitem );

				// set the panel background, and add it to the
				// card layout, with the keystring equal to
				// the text in the menu item -- this makes
				// it very easy to find from the menu item
				stats_panel.setBackground( Color.white );
				card_panel.add( cbmenuitem.getLabel(),
						stats_panel );
			}
		}

		// options menu
		options_menu = new Menu( "Options" );
		menu_bar.add( options_menu );
		options_adjWPM = new CheckboxMenuItem( "Display adjusted WPM" );
		options_adjWPM.addItemListener( this );
		options_menu.add( options_adjWPM );
		options_autorefresh = new CheckboxMenuItem(
				"Automatic refresh every second" );
		options_menu.add( options_autorefresh );

		// add the card panel to the main panel
		view_finger.setState( true );
		card_layout.show( card_panel, view_finger.getLabel() );
		add( "Center", card_panel );

		// start the refresher thread
		new Thread( this, "StatisticsGUI_refresh" ).start();
	}


	/**
	 * Thread to automatically call refresh, if the option to do it
	 * is set.
	 */
	public void run() {

		// thread 'tick', in mS
		final int 	THREAD_TICK	= 1000;

		// loop forever, at the lowest priority
		Thread.currentThread().setPriority( Thread.MIN_PRIORITY );
		while ( true ) {

			// wait for the next tick, and ignore it if our
			// window isn't visible
			do {
				try { Thread.sleep( THREAD_TICK ); }
				catch ( InterruptedException e ) {}
			} while ( ! isVisible() );

			// if the option is set, call refresh
			if ( options_autorefresh.getState() )
				refresh();
		}
	}


	/*
	 * Count how many statistics we would display on a panel, given
	 * this set and grouping - used to search for largest group.
	 */
	private int size_group_panel( KBStatistics stats_set,
			GlyphGroup group ) {

		// find the KB layout for the stats
		KBLayout kblayout = stats_set.get_kblayout();

		// now count the statistics for this group
		int num_stats = 0;
		for ( Enumeration enum = kblayout.get_elements();
				enum.hasMoreElements(); ) {
			Glyph glyph = (Glyph)enum.nextElement();

			// see if this matches the target group
			if ( glyph.get_group() == group )

				// increment the count if there are stats
				if ( stats_set.get_statistics(
						glyph.get_character() )
						!= null )
					num_stats++;
		}

		// return the count
		return ( num_stats );
	}


	/*
	 * Do the same sizing exercise for the finger panel.
	 */
	private int size_finger_panel( KBStatistics stats_set ) {

		// now count the statistics for fingers represented
		// in this stats set, starting at one because this panel
		// will also contain the global statistics
		int num_stats = 1;
		for ( Enumeration enum = Finger.get_elements();
				enum.hasMoreElements(); ) {
			Finger finger = (Finger)enum.nextElement();

			// increment the count if there are stats
			if ( stats_set.get_statistics( finger ) != null )
				num_stats++;
		}

		// return the count
		return ( num_stats );
	}


	/*
	 * Create a panel containing finger statistics.
	 */
	private Panel make_finger_stats_panel( KBStatistics stats_set,
			int stats_per_panel ) {

		// start with a set of column headings
		Panel panel = new Panel();
		int columns = COLUMNS_PER_STATISTIC * STATISTICS_PER_ROW;
		panel.setLayout( new GridLayout( 0, columns ));
		for ( int index = 0; index < STATISTICS_PER_ROW; index++ ) {
			panel.add( new Label() );
			panel.add( new Label( "Accuracy", Label.CENTER ));
			panel.add( new Label( "Speed", Label.CENTER ));
		}

		// add data labels for all the fingers we could monitor;
		// note that if a finger actually does no keys then it
		// will not appear on the stats panel
		int num_stats = 0;
		KeyingStatistics key_stats = null;
		for ( Enumeration enum = Finger.get_elements();
				enum.hasMoreElements(); ) {
			Finger finger = (Finger)enum.nextElement();

			// get finger statistics
			key_stats = stats_set.get_statistics( finger );

			// if there are any then add to the panel
			if ( key_stats != null ) {
				make_disp_entry( panel, key_stats,
					finger.get_text() );
				num_stats++;
			}
		}

		// pad the panel to the maximum size, less one for the
		// global statistics
		for ( ; num_stats < stats_per_panel - 1; num_stats ++ )
			make_disp_entry( panel, null, null );

		// add on, at the very end, the global statistics
		key_stats = stats_set.get_statistics();
		make_disp_entry( panel, key_stats, "Overall" );

		// return the panel - note how this one can't be null since
		// there must be some statistics on it!
		return ( panel );
	}


	/*
	 * Create a complete panel of labels/gauges given a char group, keymap
	 * table, and some luck.
	 */
	private Panel make_stats_panel( KBStatistics stats_set,
			GlyphGroup char_group, int stats_per_panel ) {

		// create the new panel and add headings
		Panel panel = new Panel();
		int columns = COLUMNS_PER_STATISTIC * STATISTICS_PER_ROW;
		panel.setLayout( new GridLayout( 0, columns ));
		for ( int index = 0; index < STATISTICS_PER_ROW; index++ ) {
			panel.add( new Label() );
			panel.add( new Label( "Accuracy", Label.CENTER ));
			panel.add( new Label( "Speed", Label.CENTER ));
		}

		// find the KB layout for the stats
		KBLayout kblayout = stats_set.get_kblayout();

		// now add the data gauges
		int num_stats = 0;
		for ( Enumeration enum = kblayout.get_elements();
				enum.hasMoreElements(); ) {
			Glyph glyph = (Glyph)enum.nextElement();

			// if this char is in our group then process it
			if ( glyph.get_group() == char_group ) {

				// create the stats entry for this key
				Character character = glyph.get_character();
				make_disp_entry( panel,
					stats_set.get_statistics( character ),
					glyph.get_text() );
				num_stats++;
			}
		}

		// return null if nothing on the panel, otherwise pad the
		// panel to the largest size and return it
		if ( num_stats > 0 ) {

			// pad the panel to the maximum size
			for ( ; num_stats < stats_per_panel; num_stats++ )
				make_disp_entry( panel, null, null );

			// return the panel
			return ( panel );
		}
		else
			// nothing on this one - return null
			return ( null );
	}


	/*
	 * Create a trio of components showing the specified stats
	 * set and add the trio to the panel given.
	 */
	private void make_disp_entry( Panel panel, KeyingStatistics key_stats,
			String legend ) {

		// if there is a statistics source then add the gauges
		if ( key_stats != null && legend != null ) {

			// create new gauges for this statistic
			Gauge	accuracy_gauge	= new Gauge( 0, 100, 0, "%" );
			Gauge	speed_gauge	= new Gauge( 0, 100, 0, " WPM" );

			// add a legend label and the two readout gauges
			panel.add( new Label( legend, Label.RIGHT ));
			accuracy_gauge.setBackground( Color.lightGray );
			panel.add( accuracy_gauge );
			speed_gauge.setBackground( Color.lightGray );
			panel.add( speed_gauge );

			// create a descriptor for this statistic
			StatsDescriptor descriptor = new StatsDescriptor(
					key_stats,
					accuracy_gauge, speed_gauge );
			stats_descriptors.addElement( descriptor );
		}
		else {
			// add a null entry for padding - three blank labels
			panel.add( new Label() );
			panel.add( new Label() );
			panel.add( new Label() );
		}
	}


	/**
	 * Refresh the gauges with the latest stats data.
	 */
	public void refresh() {

		// note if we are doing adjusted timings - this gets the
		// option getState out of the loop
		boolean adjusted = options_adjWPM.getState();

		// run through the list of statistics we need to update
		for ( int index = 0; index < stats_descriptors.size();
					index++ ) {

			// get the keying stats we have to use, and
			// the gauges we are going to update the
			// data into
			StatsDescriptor descriptor = (StatsDescriptor)
					stats_descriptors.elementAt( index );
			KeyingStatistics
				key_stats	= descriptor.key_stats;
			Gauge	accuracy_gauge	= descriptor.accuracy_gauge;
			Gauge	speed_gauge	= descriptor.speed_gauge;

			// if no time expended then this key or set of
			// keys has never been struck
			if ( key_stats.get_ms_expended() > 0 ) {

				// get the statistics values
				int accuracy	= key_stats.get_pc_accuracy();
				int speed	= key_stats.get_WPM( adjusted );

				// update the gauges with current values, if
				// necessary
				if ( accuracy_gauge.get_value() != accuracy ) {
					accuracy_gauge.set_value( accuracy );
					if ( accuracy <= ERROR_ACCURACY )
						accuracy_gauge.set_colour(
								Color.red );
					else if ( accuracy <= WARN_ACCURACY )
						accuracy_gauge.set_colour(
								Color.orange );
					else
						accuracy_gauge.set_colour(
								Color.black );
				}
				if ( speed_gauge.get_value() != speed ) {
					speed_gauge.set_value( speed );
					if ( speed <= ERROR_SPEED )
						speed_gauge.set_colour(
								Color.red );
					else if ( speed <= WARN_SPEED )
						speed_gauge.set_colour(
								Color.orange );
					else
						speed_gauge.set_colour(
								Color.black );
				}
			}
			else {

				// reset the gauges to nothing
				if ( accuracy_gauge.get_value()
						!= accuracy_gauge.get_min() ) {
					accuracy_gauge.set_value(
						accuracy_gauge.get_min() );
					accuracy_gauge.set_colour(
						Color.black );
				}
				if ( speed_gauge.get_value()
						!= speed_gauge.get_min() ) {
					speed_gauge.set_value(
						speed_gauge.get_min() );
					speed_gauge.set_colour(
						Color.black );
				}
			}
		}
	}


	/**
	 * Override setVisible so that the statistics get refreshed before
	 * the window is drawn.
	 */
	public void setVisible( boolean is_visible ) {
		if ( is_visible ) {

			// refresh the complete stats set before displaying
			refresh();
			super.setVisible( true );
		}
		else
			super.setVisible( false );
	}

	/*
	 * Handle dismiss, close, etc by just hiding ourselves.
	 */
	private void handle_close_button() {

		// simply just hide this frame from view
		setVisible( false );
	}


	/*
	 * Handle reset request by resetting the underlying stats
	 * then redisplaying.
	 */
	private void handle_reset_button() {
		stats_set.reset();
		refresh();
	}


	/*
	 * Handle view change by displaying the card selected.
	 */
	private void handle_view_change( CheckboxMenuItem selected ) {

		// emulate the groups thing by deselecting all menu items,
		// then reselecting just the selected one
		view_finger.setState( false );
		for ( Enumeration enum = view_glyphgroups.elements();
				enum.hasMoreElements(); ) {
			CheckboxMenuItem cbmenuitem = (CheckboxMenuItem)
					enum.nextElement();

			// unset this glyphgroup view menu item
			cbmenuitem.setState( false );
		}
		selected.setState( true );

		// display the right card panel, using the label string
		// as the key to the card
		card_layout.show( card_panel, selected.getLabel() );
		refresh();
	}

	/*
	 * Handle GUI component actions and events.
	 */
	public void actionPerformed( ActionEvent evt ) {
		if ( evt.getSource() instanceof Button ) {
			if ( evt.getSource() == refresh_button ) {
				refresh();
			}
		}
		else if ( evt.getSource() instanceof MenuItem ) {
			if ( evt.getSource() == file_close ) {
				handle_close_button();
			}
			else if ( evt.getSource() == file_refresh ) {
				refresh();
			}
			else if ( evt.getSource() == file_reset ) {
				handle_reset_button();
			}
		}
	}
	public void windowClosing( WindowEvent evt ) {
		handle_close_button();
	}
	public void windowClosed( WindowEvent evt ) {}
	public void windowOpened( WindowEvent evt ) {}
	public void windowIconified( WindowEvent evt ) {}
	public void windowDeiconified( WindowEvent evt ) {}
	public void windowActivated( WindowEvent evt ) {}
	public void windowDeactivated( WindowEvent evt ) {}
	public void itemStateChanged( ItemEvent evt ) {
		if ( evt.getSource() instanceof CheckboxMenuItem ) {
			if ( evt.getSource() == options_adjWPM )
				refresh();
			else if ( evt.getSource() != options_autorefresh )
				handle_view_change(
					(CheckboxMenuItem)evt.getSource() );
		}
	}


	/**
	 * Class test code.
	 */
	public static void main( String args[] ) {
		System.out.println( "*** CLASS TEST CODE ***" );

		KBStatistics kk = new KBStatistics( new KBLayoutUS() );
		kk.update_statistics( ' ', true, 200 );
		StatisticsGUI stats = new StatisticsGUI( false, kk );
		stats.pack(); stats.setVisible( true );
		kk.update_statistics( 'q', true, 200 );
		kk.update_statistics( 'q', true, 200 );
		kk.update_statistics( 'q', false, 50 );
	}
}
